//------------------------------------------------------------------------
//
// Author      : Maarten Kronberger
// Email       : sulacomcclaw@hotmail.com
// Website     : http://www.sulaco.co.za
// Date        : 20 December 2004
// Version     : 1.0
// Description : Tangent Bumpmapping using nVidias Cg
//               Originally done by Sren Dreijer (Halloko)
//               http://www.blacksmith-studios.dk
//               Thanx for all the help dude :)
// Change Log  :
//------------------------------------------------------------------------

program OpenGLApp;

uses
  Windows,
  Messages,
  dglOpenGL,
  Textures,
  VectorGeometry,
  SysUtils,
  cg,
  cgGL;

const
  WND_TITLE = 'Tangent Bumpmapping using nVidias Cg by McCLaw';
  FPS_TIMER = 1;                     // Timer to calculate FPS
  FPS_INTERVAL = 1000;               // Calculate FPS every 1000 ms

  // User Constants

var
  h_Wnd  : HWND;                     // Global window handle
  h_DC   : HDC;                      // Global device context
  h_RC   : HGLRC;                    // OpenGL rendering context
  keys : Array[0..255] of Boolean;   // Holds keystrokes
  FPSCount : Integer = 0;            // Counter for FPS
  ElapsedTime : Integer;             // Elapsed time between frames

  // The texture IDs for our rock texture and the normal map
  GuiRockTexture : GLuint;
  GuiNormalTexture : GLuint;

  // Mouse Variables
  MouseMove : boolean = false;        // Is the mouse moving?
  MoveXcoord, MoveYcoord : GLfloat;   // Current X and Y position of the cursor

  MouseButton : Integer = -1;         // mouse button down
  Xcoord, Ycoord : Integer;           // Current X and Y position where the mouse was clicked

  OpenGLInitialized : Boolean;

  // Quadric to generate the sphere for the light source marker 
  GlightSphere : PGLUQuadricObj;

  // The CG context
  Gcontext : CGcontext;

  // Our CG programs and their profiles
  GfragmentProgram : CGprogram;
  GfragmentProfile : CGprofile;
  GvertexProgram : CGprogram;
  GvertexProfile : CGprofile;

  // The parameter references to the vertex and fragment shader.
  // We can use the parameter references below to actually set the value of a parameter in the shaders
  GmodelViewMatrix : CGparameter;
  GlightPosition : CGparameter;
  GlightDiffuseColor : CGparameter;
  GeyePosition : CGparameter;

  // The coordinates for our quads
  GvQuads : Array [0..1] of array [0..3] of TAffineVector;

  // The texture coordinates for the 4 corners of the quads
  GvTexCoords : Array [0..1] of array [0..3] of TTexPoint;

    // Four TBNMatrices (one for each corner of the quad), and each matrix consists of 3 vector
  GTBNMatrices : Array [0..1] of Array [0..3] of Array [0..2] of TAffineVector;

  // The position of the light
  GvLightPos : TAffineVector;

  // The rotation angle of the light source (we make it go round in an infinite loop)
  GfRotAngle : Single;

  // The Speed the light will travel
  GfLightSpeed : Single;

  // A display list for our Polygons
  GDisplayList : GLuint;

  // A Display List for our Light Sphere
  GLightDisplayList : GLuint;

  {$R *.RES}

{------------------------------------------------------------------}
{  ROUNDOFF                                                        }
{------------------------------------------------------------------}
function ROUNDOFF(x : Single) : Single;
begin
  if (x < 0.0001) and (x > -0.0001) then
    Result := 0.0
  else
    Result := x;
end;

{------------------------------------------------------------------}
{  Calculate the Tangent Binomial Normal Matrices for each vertex  }
{------------------------------------------------------------------}
procedure CalculateTBNMatrix;
var
  i,j : Integer;
  nNextIndex : Integer;
  nPrevIndex : Integer;
  vVec1,
  vVec2,
  T,
  B,
  N : TAffineVector;
  s1,
  t1,
  s2,
  t2,
  fDenominator,
  fScale : Single;
begin
	// Loop through each vertex and calculate the tangent basis
  for j := 0 to 1 do
  begin
    for i := 0 to 3 do
    begin
      // Calculate the index to the right and left of the current index
      nNextIndex := (i + 1) mod 4;
      nPrevIndex := (i + 2) mod 4;

      // Calculate the vectors from the current vertex to two other corners of the quad
      vVec1 := VectorSubtract(GvQuads[j][nNextIndex],GvQuads[j][i]);
      vVec2 := VectorSubtract(GvQuads[j][nPrevIndex],GvQuads[j][i]);

      // The equation presented in the article states that:
      // s1 = V2.texcoord.x  V1.texcoord.x
      // t1 = V2.texcoord.y  V1.texcoord.y
      // s2 = V3.texcoord.x  V1.texcoord.x
      // t2 = V3.texcoord.y  V1.texcoord.y

      // Calculate s1 and t1
      s1 := GvTexCoords[j][nNextIndex].s - GvTexCoords[j][i].s;
      t1 := GvTexCoords[j][nNextIndex].t - GvTexCoords[j][i].t;

      // Calculate s2 and t2
      s2 := GvTexCoords[j][nPrevIndex].s - GvTexCoords[j][i].s;
      t2 := GvTexCoords[j][nPrevIndex].t - GvTexCoords[j][i].t;

      fDenominator := s1 * t2 - s2 * t1;
      if (ROUNDOFF(fDenominator) <= 0.0) then
      begin
        // We won't risk a divide by zero, so set the tangent matrix to the identity matrix
        GTBNMatrices[j][i][0] := AffineVectorMake(1.0, 0.0, 0.0);
        GTBNMatrices[j][i][1] := AffineVectorMake(0.0, 1.0, 0.0);
        GTBNMatrices[j][i][2] := AffineVectorMake(0.0, 0.0, 1.0);
      end
      else
      begin
        // Calculate the reciprocal value once and for all (to achieve speed)
        fScale := 1.0 / fDenominator;

        // T and B are calculated just as the equation in the article states
        T := AffineVectorMake((t2 * vVec1[0] - t1 * vVec2[0]) * fScale,
                              (t2 * vVec1[1] - t1 * vVec2[1]) * fScale,
                              (t2 * vVec1[2] - t1 * vVec2[2]) * fScale);
        T := VectorNormalize(T);

        B := AffineVectorMake((-s2 * vVec1[0] + s1 * vVec2[0]) * fScale,
                             (-s2 * vVec1[1] + s1 * vVec2[1]) * fScale,
                             (-s2 * vVec1[2] + s1 * vVec2[2]) * fScale);
        B := VectorNormalize(B);

        // The normal, N, is calculated as the cross product between T and B
        N := VectorCrossProduct(T,B);

        // Store the new tangent basis
        GTBNMatrices[j][i][0] := T;
        GTBNMatrices[j][i][1] := B;
        GTBNMatrices[j][i][2] := N;
      end;
    end;
  end;
end;


{------------------------------------------------------------------}
{  Function to Report any errors while compiling the CG            }
{------------------------------------------------------------------}
procedure CGErrorCallback; cdecl
var szError : String; // Create a buffer to hold the error string
begin

	// cgGetErrorString() returns the error that has occurred.
	// cgGetLastListing() returns a more descriptive report of the error
  szError := Format('Copy %s to %s?', [String(cgGetErrorString(cgGetError())), String(cgGetLastListing(Gcontext))]);

	// Print the error message in a standard message box
	MessageBox(0, PChar(szError), 'CG Initialization Error!', MB_OK or MB_ICONWARNING);

end;

{------------------------------------------------------------------}
{  Function to Initialize the CG programs                          }
{------------------------------------------------------------------}
function InitCG : Boolean;
begin
	// Set up the function which gets called by CG if something goes wrong
	// Well define CGErrorCallback in a second..
	cgSetErrorCallback(CGErrorCallback);

	// Create the CG context which will hold our shader programs
	Gcontext := cgCreateContext();

	// Find the best matching profile for a fragment shader
	GfragmentProfile := cgGLGetLatestProfile(CG_GL_FRAGMENT);
	cgGLSetOptimalOptions(GfragmentProfile);

	// Find the best matching profile for a vertex shader
	GvertexProfile := cgGLGetLatestProfile(CG_GL_VERTEX);
	cgGLSetOptimalOptions(GvertexProfile);

	// Create the fragment program.
	// cgCreateProgramFromFile() takes the following parameters:
	// A CG context, the type of the CG file (GL_SOURCE specifies that were reading a shader which hasnt been compiled yet), the filename of the CG 
	// shader, the shader profile we just created, the name of the start function in the shader.
	GfragmentProgram := cgCreateProgramFromFile(Gcontext, CG_SOURCE, 'FragmentShader.cg', GfragmentProfile, 'main', 0);
	if not Assigned(GfragmentProgram) then
  begin
		Result := False;
    exit;
  end;

	// Load the fragment program
	cgGLLoadProgram(GfragmentProgram);

	// Create the vertex program
	GvertexProgram := cgCreateProgramFromFile(Gcontext, CG_SOURCE, 'VertexShader.cg', GvertexProfile, 'main', 0);
	if not Assigned(GvertexProgram) then
  begin
		Result := False;
    exit;
  end;

	// Load the vertex program
	cgGLLoadProgram(GvertexProgram);

	// This calculates the TBN matrices for all vertices in the scene
	CalculateTBNMatrix();

	// Get the parameters which we can pass to the vertex and fragment shader.
	// Think of it like main(int nFoo). What we do below is to fetch a reference to the parameters (like nFoo)
	// and we can then set their value out of the CG program. Youll see the connection a bit later..
	GmodelViewMatrix := cgGetNamedParameter(GvertexProgram, 'modelViewProjMatrix');
	GlightPosition := cgGetNamedParameter(GvertexProgram, 'vLightPosition');
	GeyePosition := cgGetNamedParameter(GvertexProgram, 'vEyePosition');
	GlightDiffuseColor := cgGetNamedParameter(GfragmentProgram, 'fLightDiffuseColor');

	// Create the light sphere
	GlightSphere := gluNewQuadric();

	Result := True;
end;

{------------------------------------------------------------------}
{  Clean up memory                                                 }
{------------------------------------------------------------------}
procedure DestroyCG;
begin
	// Destroy and free our light sphere
	if (GlightSphere <> nil) then
		gluDeleteQuadric(GlightSphere);

	// Destroying the CG context automatically destroys all attached CG programs
	cgDestroyContext(Gcontext);
end;

{------------------------------------------------------------------}
{  Function to convert int to string. (No sysutils = smaller EXE)  }
{------------------------------------------------------------------}
function IntToStr(Num : Integer) : String;  // using SysUtils increase file size by 100K
begin
  Str(Num, result);
end;

{------------------------------------------------------------------}
{  Function to draw the actual scene                               }
{------------------------------------------------------------------}
procedure glDraw();
var
  i : Integer;
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);
	glLoadIdentity();


  glTranslatef(0.0,0.0,-40.0);
 
	glPushAttrib(GL_CURRENT_BIT);

    // First draw the white light source
    glColor3f(1.0, 1.0, 1.0);
    glPushMatrix();
      // We want to make the light source go round in a circle. Here we just apply some basic trigonometry.
      // Please note we're not factoring in time in this calculation (which might result in a
      // faster/slower moving circle on some computers)
      GfRotAngle := GfRotAngle + PI * 0.02 * GfLightSpeed;
      GvLightPos[0] := cos(GfRotAngle) * 20.0;
      GvLightPos[1] := sin(GfRotAngle) * 10.0;

      // Position and render the light sphere
      glTranslatef(GvLightPos[0], GvLightPos[1], GvLightPos[2]);
      if GLightDisplayList = 0 then
      begin
        GLightDisplayList := glGenLists(1);
        glNewList(GLightDisplayList,GL_COMPILE);
          gluSphere(GlightSphere, 1.0, 20, 20);
        glEndList;
      end;
      glCallList(GLightDisplayList);
    glPopMatrix();
	glPopAttrib();

	// Enable the vertex and fragment profiles and bind the vertex and fragment programs
	cgGLEnableProfile(GvertexProfile);
	cgGLEnableProfile(GfragmentProfile);

	cgGLBindProgram(GvertexProgram);
	cgGLBindProgram(GfragmentProgram);

	// Set the "modelViewProjMatrix" parameter in the vertex shader to the current concatenated
	// modelview and projection matrix
	cgGLSetStateMatrixParameter(GmodelViewMatrix, CG_GL_MODELVIEW_PROJECTION_MATRIX, CG_GL_MATRIX_IDENTITY);

	// Set the light position parameter in the vertex shader
	cgGLSetParameter3f(GlightPosition, GvLightPos[0], GvLightPos[1], GvLightPos[2]);

	// Set the diffuse of the light in the fragment shader
	cgGLSetParameter3f(GlightDiffuseColor, 1, 1, 1);

  if GDisplayList = 0 then
  begin
    GDisplayList := glGenLists(1);
    glNewList(GDisplayList,GL_COMPILE);
      // Enable and bind the rock texture
      glActiveTextureARB(GL_TEXTURE0_ARB);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, GuiRockTexture);

      // Enable and bind the normal map
      glActiveTextureARB(GL_TEXTURE1_ARB);
      glEnable(GL_TEXTURE_2D);
      glBindTexture(GL_TEXTURE_2D, GuiNormalTexture);

      glBegin(GL_QUADS);
        for i := 0 to 1 do
        begin
          // Set the texture coordinates for the rock texture and the normal map
          glMultiTexCoord2fARB(GL_TEXTURE0_ARB, GvTexCoords[i][0].s, GvTexCoords[i][0].t);

          // Specify the tangent matrix vectors, one by one, and send them to the vertex shader
          glMultiTexCoord3fARB(GL_TEXTURE1_ARB, GTBNMatrices[i][0][0][0], GTBNMatrices[i][0][0][1], GTBNMatrices[i][0][0][2]);
          glMultiTexCoord3fARB(GL_TEXTURE2_ARB, GTBNMatrices[i][0][1][0], GTBNMatrices[i][0][1][1], GTBNMatrices[i][0][1][2]);
          glMultiTexCoord3fARB(GL_TEXTURE3_ARB, GTBNMatrices[i][0][2][0], GTBNMatrices[i][0][2][1], GTBNMatrices[i][0][2][2]);

          // Draw the bottom left vertex
          glVertex3fv(@GvQuads[i][0]);

          // ------------------------------------------------------------ //

          // Set the texture coordinates for the rock texture and the normal map
          glMultiTexCoord2fARB(GL_TEXTURE0_ARB, GvTexCoords[i][1].s, GvTexCoords[i][1].t);

          // Specify the tangent matrix vectors, one by one, and send them to the vertex shader
          glMultiTexCoord3fARB(GL_TEXTURE1_ARB, GTBNMatrices[i][1][0][0], GTBNMatrices[i][1][0][1], GTBNMatrices[i][1][0][2]);
          glMultiTexCoord3fARB(GL_TEXTURE2_ARB, GTBNMatrices[i][1][1][0], GTBNMatrices[i][1][1][1], GTBNMatrices[i][1][1][2]);
          glMultiTexCoord3fARB(GL_TEXTURE3_ARB, GTBNMatrices[i][1][2][0], GTBNMatrices[i][1][2][1], GTBNMatrices[i][1][2][2]);

          // Draw the bottom right vertex
          glVertex3fv(@GvQuads[i][1]);

          // ------------------------------------------------------------ //

          // Set the texture coordinates for the rock texture and the normal map
          glMultiTexCoord2fARB(GL_TEXTURE0_ARB, GvTexCoords[i][2].s, GvTexCoords[i][2].t);

          // Specify the tangent matrix vectors, one by one, and send them to the vertex shader
          glMultiTexCoord3fARB(GL_TEXTURE1_ARB, GTBNMatrices[i][2][0][0], GTBNMatrices[i][2][0][1], GTBNMatrices[i][2][0][2]);
          glMultiTexCoord3fARB(GL_TEXTURE2_ARB, GTBNMatrices[i][2][1][0], GTBNMatrices[i][2][1][1], GTBNMatrices[i][2][1][2]);
          glMultiTexCoord3fARB(GL_TEXTURE3_ARB, GTBNMatrices[i][2][2][0], GTBNMatrices[i][2][2][1], GTBNMatrices[i][2][2][2]);

          // Draw the top right vertex
          glVertex3fv(@GvQuads[i][2]);

          // ------------------------------------------------------------ //

          // Set the texture coordinates for the rock texture and the normal map
          glMultiTexCoord2fARB(GL_TEXTURE0_ARB, GvTexCoords[i][3].s, GvTexCoords[i][3].t);

          // Specify the tangent matrix vectors, one by one, and send them to the vertex shader
          glMultiTexCoord3fARB(GL_TEXTURE1_ARB, GTBNMatrices[i][3][0][0], GTBNMatrices[i][3][0][1], GTBNMatrices[i][3][0][2]);
          glMultiTexCoord3fARB(GL_TEXTURE2_ARB, GTBNMatrices[i][3][1][0], GTBNMatrices[i][3][1][1], GTBNMatrices[i][3][1][2]);
          glMultiTexCoord3fARB(GL_TEXTURE3_ARB, GTBNMatrices[i][3][2][0], GTBNMatrices[i][3][2][1], GTBNMatrices[i][3][2][2]);

          // Draw the top left vertex
          glVertex3fv(@GvQuads[i][3]);
        end;
      glEnd();

      // Disable textures
      glActiveTextureARB(GL_TEXTURE0_ARB);
      glDisable(GL_TEXTURE_2D);

      glActiveTextureARB(GL_TEXTURE1_ARB);
      glDisable(GL_TEXTURE_2D);

    glEndList();
  end;

  glCallList(GDisplayList);

	// Disable the vertex and fragment profiles again
	cgGLDisableProfile(GvertexProfile);
	cgGLDisableProfile(GfragmentProfile);

	
  // Flush the OpenGL Buffer
  glFlush();                            // ( Force the buffer to draw or send a network packet of commands in a networked system)

end;


{------------------------------------------------------------------}
{  Initialise OpenGL                                               }
{------------------------------------------------------------------}
procedure glInit();
begin
  glClearColor(0.0, 0.0, 0.0, 0.0); 	   // Black Background
  glShadeModel(GL_SMOOTH);                 // Enables Smooth Color Shading
  glClearDepth(1.0);                       // Depth Buffer Setup
  glEnable(GL_DEPTH_TEST);                 // Enable Depth Buffer
  glDepthFunc(GL_LESS);		           // The Type Of Depth Test To Do

  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);   //Realy Nice perspective calculations
  //glEnable(GL_CULL_FACE);
  glEnable(GL_TEXTURE_2D);                     // Enable Texture Mapping
  LoadTexture('Rock.bmp', GuiRockTexture, FALSE);    // Load the Texture
  LoadTexture('RockNormal.bmp', GuiNormalTexture, FALSE);    // Load the Texture

  //LoadTexture('texture.bmp', GuiRockTexture, FALSE);    // Load the Texture
  //LoadTexture('nrm.bmp', GuiNormalTexture, FALSE);    // Load the Texture

  GvQuads[0][0] := AffineVectorMake(-20.0, -10.0, -0.0);
  GvQuads[0][1] := AffineVectorMake( 0.0, -10.0, -0.0);
  GvQuads[0][2] := AffineVectorMake( 0.0, 10.0, -0.0);
  GvQuads[0][3] := AffineVectorMake(-20.0, 10.0, -0.0);
  GvQuads[1][0] := AffineVectorMake(0.0, -10.0, -0.0);
  GvQuads[1][1] := AffineVectorMake(20.0, -10.0, -20.0);
  GvQuads[1][2] := AffineVectorMake(20.0, 10.0, -20.0);
  GvQuads[1][3] := AffineVectorMake(0.0, 10.0, -0.0);

  // The texture coordinates for the 4 corners of the quad
  GvTexCoords[0][0] := TexPointMake(0.0, 0.0);
  GvTexCoords[0][1] := TexPointMake(1.0, 0.0);
  GvTexCoords[0][2] := TexPointMake(1.0, 1.0);
  GvTexCoords[0][3] := TexPointMake(0.0, 1.0);
  GvTexCoords[1][0] := TexPointMake(0.0, 0.0);
  GvTexCoords[1][1] := TexPointMake(1.0, 0.0);
  GvTexCoords[1][2] := TexPointMake(1.0, 1.0);
  GvTexCoords[1][3] := TexPointMake(0.0, 1.0);
  // The position of the light
  GvLightPos := AffineVectorMake(0.0, 0.0, 10.0);

  // The rotation angle of the light source (we make it go round in an infinite loop)
  GfRotAngle := 0.0;
  GfLightSpeed := 0.1;

  InitCG;

  GDisplayList := 0;
end;


{------------------------------------------------------------------}
{  Handle window resize                                            }
{------------------------------------------------------------------}
procedure glResizeWnd(Width, Height : Integer);
begin
  if (Height = 0) then                // prevent divide by zero exception
    Height := 1;
  glViewport(0, 0, Width, Height);    // Set the viewport for the OpenGL window
  glMatrixMode(GL_PROJECTION);        // Change Matrix Mode to Projection
  glLoadIdentity();                   // Reset View
  gluPerspective(45.0, Width/Height, 1.0, 100.0);  // Do the perspective calculations. Last value = max clipping depth

  glMatrixMode(GL_MODELVIEW);         // Return to the modelview matrix
  glLoadIdentity();                   // Reset View
end;

{------------------------------------------------------------------}
{  Processes all the mouse move                                  }
{------------------------------------------------------------------}
procedure ProcessMouseMove();
begin
  if MouseMove then
  begin
    // Add Code here to execute on mouse move
    MouseMove := False;
  end;
end;

{------------------------------------------------------------------}
{  Processes all the mouse clicks                                  }
{------------------------------------------------------------------}
procedure ProcessMouseClick;
begin
  case MouseButton of
  1: // Left Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  2: // Right Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  3: // Middle Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0;  // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  end;
end;

{------------------------------------------------------------------}
{  Processes all the keystrokes                                    }
{------------------------------------------------------------------}
procedure ProcessKeys;
begin

end;

{------------------------------------------------------------------}
{  Determines the applications response to the messages received  }
{------------------------------------------------------------------}
function WndProc(hWnd: HWND; Msg: UINT;  wParam: WPARAM;  lParam: LPARAM): LRESULT; stdcall;
begin
  case (Msg) of
    WM_CREATE:
      begin
        // Insert stuff you want executed when the program starts
      end;
    WM_CLOSE:
      begin
        PostQuitMessage(0);
        Result := 0
      end;
    WM_KEYDOWN:       // Set the pressed key (wparam) to equal true so we can check if its pressed
      begin
        keys[wParam] := True;
        Result := 0;
      end;
    WM_KEYUP:         // Set the released key (wparam) to equal false so we can check if its pressed
      begin
        keys[wParam] := False;
        Result := 0;
      end;
    WM_MOUSEMOVE:
      begin
        MouseMove := True;
        MoveXcoord := LOWORD(lParam);
        MoveYcoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 1;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_RBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 2;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_MBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 3;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONUP,WM_RBUTTONUP,WM_MBUTTONUP:
      begin
        ReleaseCapture();   // above
        MouseButton := 0;
        XCoord := 0;
        YCoord := 0;
        Result := 0;
      end;
    WM_SIZE:          // Resize the window with the new width and height
      begin
        if OpenGLInitialized then
         glResizeWnd(LOWORD(lParam),HIWORD(lParam));
        Result := 0;
      end;
    WM_TIMER :                     // Add code here for all timers to be used.
      begin
        if wParam = FPS_TIMER then
        begin
          FPSCount :=Round(FPSCount * 1000/FPS_INTERVAL);   // calculate to get per Second incase intercal is less or greater than 1 second
          SetWindowText(h_Wnd, PChar(WND_TITLE + '   [' + intToStr(FPSCount) + ' FPS]'));
          FPSCount := 0;
          Result := 0;
        end;
      end;
    else
      begin
      
      Result := DefWindowProc(hWnd, Msg, wParam, lParam);    // Default result if nothing happens
      end;
  end;
end;


{---------------------------------------------------------------------}
{  Properly destroys the window created at startup (no memory leaks)  }
{---------------------------------------------------------------------}
procedure glKillWnd(Fullscreen : Boolean);
begin
  if Fullscreen then             // Change back to non fullscreen
  begin
    ChangeDisplaySettings(devmode(nil^), 0);
    ShowCursor(True);
  end;
  
  // Makes current rendering context not current, and releases the device
  // context that is used by the rendering context.
  if (not wglMakeCurrent(h_DC, 0)) then
    MessageBox(0, 'Release of DC and RC failed!', 'Error', MB_OK or MB_ICONERROR);

  // Attempts to delete the rendering context
  if (not wglDeleteContext(h_RC)) then
  begin
    MessageBox(0, 'Release of rendering context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_RC := 0;
  end;

  // Attemps to release the device context
  if ((h_DC > 0) and (ReleaseDC(h_Wnd, h_DC) = 0)) then
  begin
    MessageBox(0, 'Release of device context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_DC := 0;
  end;

  // Attempts to destroy the window
  if ((h_Wnd <> 0) and (not DestroyWindow(h_Wnd))) then
  begin
    MessageBox(0, 'Unable to destroy window!', 'Error', MB_OK or MB_ICONERROR);
    h_Wnd := 0;
  end;



  // Attempts to unregister the window class
  if (not UnRegisterClass('OpenGL', hInstance)) then
  begin
    MessageBox(0, 'Unable to unregister window class!', 'Error', MB_OK or MB_ICONERROR);
    hInstance := 0;
  end;
end;


{--------------------------------------------------------------------}
{  Creates the window and attaches a OpenGL rendering context to it  }
{--------------------------------------------------------------------}
function glCreateWnd(Width, Height : Integer; Fullscreen : Boolean; PixelDepth : Integer) : Boolean;
var
  wndClass : TWndClass;         // Window class
  dwStyle : DWORD;              // Window styles
  dwExStyle : DWORD;            // Extended window styles
  dmScreenSettings : DEVMODE;   // Screen settings (fullscreen, etc...)
  PixelFormat : GLuint;         // Settings for the OpenGL rendering
  h_Instance : HINST;           // Current instance
  pfd : TPIXELFORMATDESCRIPTOR;  // Settings for the OpenGL window
begin
  InitOpenGL;                               // New call to initialize and bind the OpenGL dll

  h_Instance := GetModuleHandle(nil);       //Grab An Instance For Our Window
  ZeroMemory(@wndClass, SizeOf(wndClass));  // Clear the window class structure

  with wndClass do                    // Set up the window class
  begin
    style         := CS_HREDRAW or    // Redraws entire window if length changes
                     CS_VREDRAW or    // Redraws entire window if height changes
                     CS_OWNDC;        // Unique device context for the window
    lpfnWndProc   := @WndProc;        // Set the window procedure to our func WndProc
    hInstance     := h_Instance;
    hCursor       := LoadCursor(0, IDC_ARROW);
    lpszClassName := 'OpenGL';
  end;

  if (RegisterClass(wndClass) = 0) then  // Attemp to register the window class
  begin
    MessageBox(0, 'Failed to register the window class!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit
  end;

  // Change to fullscreen if so desired
  if Fullscreen then
  begin
    ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));
    with dmScreenSettings do begin              // Set parameters for the screen setting
      dmSize       := SizeOf(dmScreenSettings);
      dmPelsWidth  := Width;                    // Window width
      dmPelsHeight := Height;                   // Window height
      dmBitsPerPel := PixelDepth;               // Window color depth
      dmFields     := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
    end;

    // Try to change screen mode to fullscreen
    if (ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED) then
    begin
      MessageBox(0, 'Unable to switch to fullscreen!', 'Error', MB_OK or MB_ICONERROR);
      Fullscreen := False;
    end;
  end;

  // If we are still in fullscreen then
  if (Fullscreen) then
  begin
    dwStyle := WS_POPUP or                // Creates a popup window
               WS_CLIPCHILDREN            // Doesn't draw within child windows
               or WS_CLIPSIBLINGS;        // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW;         // Top level window
    ShowCursor(False);                    // Turn of the cursor (gets in the way)
  end
  else
  begin
    dwStyle := WS_OVERLAPPEDWINDOW or     // Creates an overlapping window
               WS_CLIPCHILDREN or         // Doesn't draw within child windows
               WS_CLIPSIBLINGS;           // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW or       // Top level window
                 WS_EX_WINDOWEDGE;        // Border with a raised edge
  end;

  // Attempt to create the actual window
  h_Wnd := CreateWindowEx(dwExStyle,      // Extended window styles
                          'OpenGL',       // Class name
                          WND_TITLE,      // Window title (caption)
                          dwStyle,        // Window styles
                          0, 0,           // Window position
                          Width, Height,  // Size of window
                          0,              // No parent window
                          0,              // No menu
                          h_Instance,     // Instance
                          nil);           // Pass nothing to WM_CREATE

  if h_Wnd = 0 then
  begin
    glKillWnd(Fullscreen);                // Undo all the settings we've changed
    MessageBox(0, 'Unable to create window!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Try to get a device context
  h_DC := GetDC(h_Wnd);
  if (h_DC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to get a device context!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Settings for the OpenGL window
  with pfd do
  begin
    nSize           := SizeOf(TPIXELFORMATDESCRIPTOR); // Size Of This Pixel Format Descriptor
    nVersion        := 1;                    // The version of this data structure
    dwFlags         := PFD_DRAW_TO_WINDOW    // Buffer supports drawing to window
                       or PFD_SUPPORT_OPENGL // Buffer supports OpenGL drawing
                       or PFD_DOUBLEBUFFER;  // Supports double buffering
    iPixelType      := PFD_TYPE_RGBA;        // RGBA color format
    cColorBits      := PixelDepth;           // OpenGL color depth
    cRedBits        := 0;                    // Number of red bitplanes
    cRedShift       := 0;                    // Shift count for red bitplanes
    cGreenBits      := 0;                    // Number of green bitplanes
    cGreenShift     := 0;                    // Shift count for green bitplanes
    cBlueBits       := 0;                    // Number of blue bitplanes
    cBlueShift      := 0;                    // Shift count for blue bitplanes
    cAlphaBits      := 0;                    // Not supported
    cAlphaShift     := 0;                    // Not supported
    cAccumBits      := 0;                    // No accumulation buffer
    cAccumRedBits   := 0;                    // Number of red bits in a-buffer
    cAccumGreenBits := 0;                    // Number of green bits in a-buffer
    cAccumBlueBits  := 0;                    // Number of blue bits in a-buffer
    cAccumAlphaBits := 0;                    // Number of alpha bits in a-buffer
    cDepthBits      := 16;                   // Specifies the depth of the depth buffer
    cStencilBits    := 0;                    // Turn off stencil buffer
    cAuxBuffers     := 0;                    // Not supported
    iLayerType      := PFD_MAIN_PLANE;       // Ignored
    bReserved       := 0;                    // Number of overlay and underlay planes
    dwLayerMask     := 0;                    // Ignored
    dwVisibleMask   := 0;                    // Transparent color of underlay plane
    dwDamageMask    := 0;                     // Ignored
  end;

  // Attempts to find the pixel format supported by a device context that is the best match to a given pixel format specification.
  PixelFormat := ChoosePixelFormat(h_DC, @pfd);
  if (PixelFormat = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to find a suitable pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Sets the specified device context's pixel format to the format specified by the PixelFormat.
  if (not SetPixelFormat(h_DC, PixelFormat, @pfd)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to set the pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Create a OpenGL rendering context
  h_RC := wglCreateContext(h_DC);
  if (h_RC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to create an OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Makes the specified OpenGL rendering context the calling thread's current rendering context
  if (not wglMakeCurrent(h_DC, h_RC)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to activate OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;
  // Read And Assign Extentions
  ReadExtensions;
  ReadImplementationProperties;
  OpenGLInitialized := True;
  // Initializes the timer used to calculate the FPS
  SetTimer(h_Wnd, FPS_TIMER, FPS_INTERVAL, nil);

  // Settings to ensure that the window is the topmost window
  ShowWindow(h_Wnd, SW_SHOW);
  SetForegroundWindow(h_Wnd);
  SetFocus(h_Wnd);

  // Ensure the OpenGL window is resized properly
  glResizeWnd(Width, Height);
  glInit();

  Result := True;
end;


{--------------------------------------------------------------------}
{  Main message loop for the application                             }
{--------------------------------------------------------------------}
function WinMain(hInstance : HINST; hPrevInstance : HINST;
                 lpCmdLine : PChar; nCmdShow : Integer) : Integer; stdcall;
var
  msg : TMsg;
  finished : Boolean;
  DemoStart, LastTime : DWord;
begin
  finished := False;

  // Perform application initialization:
  if not glCreateWnd(800, 600, false, 32) then
  begin
    Result := 0;
    Exit;
  end;

  DemoStart := GetTickCount();            // Get Time when demo started

  // Main message loop:
  while not finished do
  begin
    if (PeekMessage(msg, 0, 0, 0, PM_REMOVE)) then // Check if there is a message for this window
    begin
      if (msg.message = WM_QUIT) then     // If WM_QUIT message received then we are done
        finished := True
      else
      begin                               // Else translate and dispatch the message to this window
  	    TranslateMessage(msg);
        DispatchMessage(msg);
      end;
    end
    else
    begin
      Inc(FPSCount);                      // Increment FPS Counter

      LastTime := ElapsedTime;
      ElapsedTime :=GetTickCount() - DemoStart;     // Calculate Elapsed Time
      ElapsedTime :=(LastTime + ElapsedTime) DIV 2; // Average it out for smoother movement

      glDraw();                           // Draw the scene
      SwapBuffers(h_DC);                  // Display the scene

      if (keys[VK_ESCAPE]) then           // If user pressed ESC then set finised TRUE
        finished := True
      else
      begin
        ProcessKeys;                      // Check for any other key Pressed
        ProcessMouseMove;                 // Check for mouse move
        ProcessMouseClick;                // Check for Mouse Click
      end;
    end;
  end;
  DestroyCG;
  glKillWnd(FALSE);
  Result := msg.wParam;
end;


begin
  WinMain( hInstance, hPrevInst, CmdLine, CmdShow );
end.
